"""
NPlib - A library for numerical processing.

This package provides various numerical processing utilities.
"""

__version__ = "0.1.0"
__author__ = "Riccardo Farris"
__credits__ = 'Reactive Nanoparticles Laboratory'
